import random
from datetime import datetime


class DummyRequestStatement:
    '''A mock scheduler for tests.

    >>> async def main():
    ...     scheduler = DummyRequestStatement()
    ...     return await scheduler()

    >>> import asyncio
    >>> script = asyncio.run(main())
    >>> code = compile(script, '<string>', 'exec')
    >>> # exec(code, globals())

    '''

    async def __call__(self) -> str:
        return generate_dummy_statement()


def generate_dummy_statement() -> str:
    s = '\n'.join(
        f'time.sleep({random.uniform(1, 3):.2})' for _ in range(random.randint(5, 10))
    )
    s = '\n'.join(
        (
            f'# {datetime.utcnow().isoformat()}',
            'import time',
            '',
            s,
        )
    )
    return s



import strawberry
from strawberry.types import Info

from nextline_schedule.queue import PushArg, Queue

from .types import ScheduleQueueItem


@strawberry.input
class ScheduleQueuePushInput:
    name: str
    script: str

    def to_push_arg(self) -> PushArg:
        return PushArg(
            name=self.name,
            script=self.script,
        )


async def mutate_push(info: Info, input: ScheduleQueuePushInput) -> ScheduleQueueItem:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, Queue)
    push_arg = input.to_push_arg()
    item = await queue.push(push_arg)
    return ScheduleQueueItem.from_(item)


async def mutate_remove(info: Info, id: int) -> bool:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, Queue)
    return await queue.remove(id)


async def mutate_move_to_first(info: Info, id: int) -> bool:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, Queue)
    return await queue.move_to_first(id)


async def mutate_move_to_last(info: Info, id: int) -> bool:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, Queue)
    return await queue.move_to_last(id)


async def mutate_move_one_forward(info: Info, id: int) -> bool:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, Queue)
    return await queue.move_one_forward(id)


async def mutate_move_one_backward(info: Info, id: int) -> bool:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, Queue)
    return await queue.move_one_backward(id)


@strawberry.type
class MutationScheduleQueueMove:
    to_first: bool = strawberry.mutation(resolver=mutate_move_to_first)
    to_last: bool = strawberry.mutation(resolver=mutate_move_to_last)
    one_forward: bool = strawberry.mutation(resolver=mutate_move_one_forward)
    one_backward: bool = strawberry.mutation(resolver=mutate_move_one_backward)


@strawberry.type
class MutationScheduleQueue:
    push: ScheduleQueueItem = strawberry.mutation(resolver=mutate_push)
    remove: bool = strawberry.mutation(resolver=mutate_remove)

    @strawberry.mutation
    def move(self) -> MutationScheduleQueueMove:
        return MutationScheduleQueueMove()



from unittest.mock import Mock

from deepdiff import DeepDiff
from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.graphql import MUTATE_QUEUE_MOVE_TO_LAST
from nextline_schedule.queue import Queue
from nextline_test_utils.strategies import st_graphql_ints
from tests.schema.conftest import Schema


@given(d=st.data())
async def test_schema(d: st.DataObject, schema: Schema) -> None:
    id = d.draw(st_graphql_ints(min_value=1))
    return_value = d.draw(st.booleans())

    queue = Mock(spec=Queue)
    queue.move_to_last.return_value = return_value

    context_values = {'schedule': {'queue': queue}}
    variable_values = {'id': id}

    resp = await schema.execute(
        MUTATE_QUEUE_MOVE_TO_LAST,
        context_value=context_values,
        variable_values=variable_values,
    )

    assert (data := resp.data)
    expected_data = {'schedule': {'queue': {'move': {'toLast': return_value}}}}
    assert DeepDiff(data, expected_data) == {}

    queue.move_to_last.assert_awaited_once_with(id)



from os import PathLike
from pathlib import Path

from graphql import parse, print_ast


def read_gql(path: PathLike | str) -> str:
    '''Load a GraphQL query from a file while checking its syntax.'''

    text = Path(path).read_text()
    parsed = parse(text)
    reformatted = print_ast(parsed)
    return reformatted


pwd = Path(__file__).resolve().parent

sub = pwd / 'mutations'
MUTATE_AUTO_MODE_TURN_ON = read_gql(sub / 'auto_mode' / 'TurnOn.gql')
MUTATE_AUTO_MODE_TURN_OFF = read_gql(sub / 'auto_mode' / 'TurnOff.gql')
MUTATE_AUTO_MODE_CHANGE_MODE = read_gql(sub / 'auto_mode' / 'ChangeMode.gql')
MUTATE_QUEUE_PUSH = read_gql(sub / 'queue' / 'Push.gql')
MUTATE_QUEUE_REMOVE = read_gql(sub / 'queue' / 'Remove.gql')
MUTATE_QUEUE_MOVE_ONE_BACKWARD = read_gql(sub / 'queue' / 'move' / 'OneBackward.gql')
MUTATE_QUEUE_MOVE_ONE_FORWARD = read_gql(sub / 'queue' / 'move' / 'OneForward.gql')
MUTATE_QUEUE_MOVE_TO_FIRST = read_gql(sub / 'queue' / 'move' / 'ToFirst.gql')
MUTATE_QUEUE_MOVE_TO_LAST = read_gql(sub / 'queue' / 'move' / 'ToLast.gql')
MUTATE_SCHEDULER_UPDATE = read_gql(sub / 'scheduler' / 'Update.gql')
MUTATE_SCHEDULER_LOAD_SCRIPT = read_gql(sub / 'scheduler' / 'LoadScript.gql')

sub = pwd / 'queries'
QUERY_VERSION = read_gql(sub / 'Version.gql')
QUERY_AUTO_MODE_STATE = read_gql(sub / 'auto_mode' / 'State.gql')
QUERY_AUTO_MODE_MODE = read_gql(sub / 'auto_mode' / 'Mode.gql')
QUERY_QUEUE_ITEMS = read_gql(sub / 'queue' / 'Items.gql')
QUERY_SCHEDULER = read_gql(sub / 'scheduler' / 'Scheduler.gql')
QUERY_SCHEDULER_PREVIEW = read_gql(sub / 'scheduler' / 'Preview.gql')

sub = pwd / 'subscriptions'
SUBSCRIBE_AUTO_MODE_STATE = read_gql(sub / 'ScheduleAutoModeState.gql')
SUBSCRIBE_AUTO_MODE_MODE = read_gql(sub / 'ScheduleAutoModeMode.gql')
SUBSCRIBE_QUEUE_ITEMS = read_gql(sub / 'ScheduleQueueItems.gql')



from string import ascii_letters
from typing import TypedDict
from unittest.mock import Mock

from deepdiff import DeepDiff
from hypothesis import Phase, given, provisional, settings
from hypothesis import strategies as st

from nextline_schedule.graphql import MUTATE_SCHEDULER_UPDATE
from nextline_schedule.scheduler import Scheduler
from nextline_test_utils.strategies import st_graphql_ints, st_none_or
from tests.schema.conftest import Schema


class Input(TypedDict):
    apiUrl: str | None
    lengthMinutes: int | None
    policy: str | None


@st.composite
def st_input(draw: st.DrawFn) -> Input:
    apiUrl = draw(st_none_or(provisional.urls()))
    lengthMinutes = draw(st_none_or(st_graphql_ints(min_value=1)))
    policy = draw(st_none_or(st.text(ascii_letters, min_size=1)))
    return {'apiUrl': apiUrl, 'lengthMinutes': lengthMinutes, 'policy': policy}


@settings(phases=(Phase.generate,))  # To avoid shrinking
@given(d=st.data())
async def test_schema(d: st.DataObject, schema: Schema) -> None:
    input = d.draw(st_input())

    init_api_url = d.draw(provisional.urls())
    init_length_minutes = d.draw(st_graphql_ints(min_value=1))
    init_policy = d.draw(st.text(ascii_letters, min_size=1))

    scheduler = Mock(spec=Scheduler)
    scheduler._api_url = init_api_url
    scheduler._length_minutes = init_length_minutes
    scheduler._policy = init_policy

    context_values = {'schedule': {'scheduler': scheduler}}
    variable_values = {'input': input}

    resp = await schema.execute(
        MUTATE_SCHEDULER_UPDATE,
        context_value=context_values,
        variable_values=variable_values,
    )

    assert (data := resp.data)
    expected_data = {'schedule': {'scheduler': {'update': True}}}
    assert DeepDiff(data, expected_data) == {}

    if input['apiUrl'] is not None:
        assert scheduler._api_url == input['apiUrl']
    else:
        assert scheduler._api_url == init_api_url

    if input['lengthMinutes'] is not None:
        assert scheduler._length_minutes == input['lengthMinutes']
    else:
        assert scheduler._length_minutes == init_length_minutes

    if input['policy'] is not None:
        assert scheduler._policy == input['policy']
    else:
        assert scheduler._policy == init_policy



from strawberry.types import ExecutionResult

import nextline_schedule
from nextline_schedule.graphql import QUERY_VERSION
from tests.schema.conftest import Schema


async def test_schema(schema: Schema) -> None:
    resp = await schema.execute(QUERY_VERSION)
    assert isinstance(resp, ExecutionResult)
    assert resp.data
    assert resp.data['schedule']['version'] == nextline_schedule.__version__



from typing import Optional

import strawberry
from strawberry.types import Info

from nextline import Nextline
from nextline_schedule.scheduler import Scheduler


@strawberry.input
class MutationSchedulerUpdateInput:
    api_url: Optional[str] = None
    length_minutes: Optional[int] = None
    policy: Optional[str] = None


def mutate_update(info: Info, input: MutationSchedulerUpdateInput) -> bool:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, Scheduler)
    if input.api_url is not None:
        scheduler._api_url = input.api_url
    if input.length_minutes is not None:
        scheduler._length_minutes = input.length_minutes
    if input.policy is not None:
        scheduler._policy = input.policy
    return True


async def mutate_load_script(info: Info) -> bool:
    nextline = info.context["nextline"]
    assert isinstance(nextline, Nextline)
    scheduler = info.context['schedule']['scheduler']
    assert callable(scheduler)
    statement = await scheduler()
    assert isinstance(statement, str)
    await nextline.reset(statement=statement)
    return True


@strawberry.type
class MutationScheduleScheduler:
    update: bool = strawberry.mutation(resolver=mutate_update)
    load_script: bool = strawberry.mutation(resolver=mutate_load_script)



import asyncio
import time
from collections.abc import Callable
from typing import NoReturn

from nextline import Nextline
from nextline_schedule.auto import AutoMode

from .funcs import pull_func_factory


class MockError(Exception):
    pass


def f() -> None:  # pragma: no cover
    time.sleep(0.001)


def g() -> NoReturn:  # pragma: no cover
    time.sleep(0.001)
    raise MockError()


async def test_on_raised_while_pulling() -> None:
    async def pull() -> NoReturn:
        raise MockError()

    nextline = Nextline(statement=f, timeout_on_exit=60)
    queue = pull_func_factory('queue')
    auto_mode = AutoMode(nextline=nextline, scheduler=pull, queue=queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await auto_mode.turn_on()
            async for state in auto_mode.subscribe_state():
                if state == 'off':  # pragma: no branch
                    break

    expected = ['off', 'auto_waiting', 'auto_pulling', 'off']
    assert expected == await states


async def test_on_raised_while_running() -> None:
    async def pull() -> Callable[[], None]:
        return g

    nextline = Nextline(statement=f, timeout_on_exit=60)
    queue = pull_func_factory('queue')
    auto_mode = AutoMode(nextline=nextline, scheduler=pull, queue=queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await auto_mode.turn_on()
            async for state in auto_mode.subscribe_state():
                if state == 'off':
                    break

    expected = ['off', 'auto_waiting', 'auto_pulling', 'auto_running', 'off']
    assert expected == await states


async def subscribe_state(auto_mode: AutoMode) -> list[str]:
    return [state async for state in auto_mode.subscribe_state()]



import strawberry

from nextline_schedule import __version__

from .auto import QueryScheduleAutoMode
from .queue import QueryScheduleQueue
from .scheduler import QueryScheduleScheduler


@strawberry.type
class QuerySchedule:
    version: str = __version__

    @strawberry.field
    def auto_mode(self) -> QueryScheduleAutoMode:
        return QueryScheduleAutoMode()

    @strawberry.field
    def scheduler(self) -> QueryScheduleScheduler:
        return QueryScheduleScheduler()

    @strawberry.field
    def queue(self) -> QueryScheduleQueue:
        return QueryScheduleQueue()


@strawberry.type
class Query:
    @strawberry.field
    def schedule(self) -> QuerySchedule:
        return QuerySchedule()



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

from hypothesis import Phase, given, settings
from hypothesis import strategies as st

from nextline import Nextline
from nextline.utils import until_true
from nextline_schedule.auto import AutoMode, ModeName

from .funcs import STATEMENT_TEMPLATE, pull_func_factory, until_state


class StatefulTest:
    def __init__(self, data: st.DataObject) -> None:
        self._draw = data.draw
        statement = STATEMENT_TEMPLATE.format(name='init', count=1)
        self._nextline = Nextline(statement=statement)
        self._scheduler = pull_func_factory('schedule')
        self._queue = pull_func_factory('queue')
        self._auto_mode = AutoMode(
            nextline=self._nextline, scheduler=self._scheduler, queue=self._queue
        )

        assert self._auto_mode.mode == 'off'
        assert self._auto_mode.state == 'created'

    @asynccontextmanager
    async def context(self) -> AsyncIterator[None]:
        await asyncio.sleep(0)
        yield

    async def _subscribe_mode(self) -> list[str]:
        return [i async for i in self._auto_mode.subscribe_mode()]

    async def _subscribe_state(self) -> list[str]:
        return [i async for i in self._auto_mode.subscribe_state()]

    async def change_mode(self) -> None:
        MODE_NAMES: list[ModeName] = ['off', 'scheduler', 'queue']
        mode = self._draw(st.sampled_from(MODE_NAMES))
        await self._auto_mode.change_mode(mode)
        if mode == 'off':
            await until_true(lambda: self._auto_mode.mode == 'off')
            await until_true(lambda: self._auto_mode.state == 'off')
        elif mode == 'scheduler':
            await until_true(lambda: self._auto_mode.mode == 'scheduler')
            await until_true(lambda: self._auto_mode.state.startswith('auto_'))
        elif mode == 'queue':  # pragma: no branch
            await until_true(lambda: self._auto_mode.mode == 'queue')
            await until_true(lambda: self._auto_mode.state.startswith('auto_'))

    async def turn_on(self) -> None:
        await self._auto_mode.turn_on()
        await until_true(lambda: self._auto_mode.mode in ('scheduler', 'queue'))
        await until_true(lambda: self._auto_mode.state.startswith('auto_'))

    async def turn_off(self) -> None:
        await self._auto_mode.turn_off()
        await until_true(lambda: self._auto_mode.mode == 'off')
        await until_true(lambda: self._auto_mode.state == 'off')

    async def wait(self) -> None:
        if not self._auto_mode.state.startswith('auto_'):
            return
        state = self._draw(st.sampled_from(['finished', 'running']))
        count = self._draw(st.integers(min_value=1, max_value=3))

        await until_state(self._nextline, state=state, count=count)

    async def __aenter__(self) -> 'StatefulTest':
        self._mode_sub = asyncio.create_task(self._subscribe_mode())
        self._state_sub = asyncio.create_task(self._subscribe_state())

        await asyncio.sleep(0)

        await self._auto_mode.__aenter__()
        await self._nextline.__aenter__()

        assert self._auto_mode.mode == 'off'
        assert self._auto_mode.state == 'off'

        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self._nextline.__aexit__(*args, **kwargs)
        await self._auto_mode.__aexit__(*args, **kwargs)

        modes = await self._mode_sub
        states = await self._state_sub

        # assert modes == ['off']
        # assert states == ['off']


# @settings(max_examples=500, phases=(Phase.generate,))
@settings(deadline=None, phases=(Phase.generate,))
@given(data=st.data())
async def test_property(data: st.DataObject) -> None:
    test = StatefulTest(data)

    METHODS = [
        test.change_mode,
        test.turn_on,
        test.turn_off,
        test.wait,
    ]

    methods = data.draw(st.lists(st.sampled_from(METHODS), max_size=4))

    async with test:
        for method in methods:
            async with test.context():
                await method()



from unittest.mock import AsyncMock

from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.graphql import QUERY_SCHEDULER_PREVIEW
from nextline_test_utils.strategies import st_python_scripts
from tests.schema.conftest import Schema


@given(d=st.data())
async def test_schema(d: st.DataObject, schema: Schema) -> None:
    script = d.draw(st_python_scripts())
    scheduler = AsyncMock(return_value=script)

    context = {'schedule': {'scheduler': scheduler}}
    resp = await schema.execute(QUERY_SCHEDULER_PREVIEW, context_value=context)
    assert (data := resp.data)

    expected = {
        'schedule': {
            'scheduler': {
                'preview': {
                    'script': script,
                }
            }
        }
    }
    assert data == expected



from copy import deepcopy
from pathlib import Path
from unittest.mock import AsyncMock

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st
from transitions import Machine
from transitions.extensions import HierarchicalAsyncGraphMachine
from transitions.extensions.asyncio import HierarchicalAsyncMachine
from transitions.extensions.markup import HierarchicalMarkupMachine

from nextline_schedule.auto.state_machine.config import CONFIG

SELF_LITERAL = Machine.self_literal


def test_model_default() -> None:
    machine = HierarchicalAsyncMachine(model=None, **CONFIG)  # type: ignore
    assert not machine.models


def test_model_self_literal() -> None:
    machine = HierarchicalAsyncMachine(model=SELF_LITERAL, **CONFIG)  # type: ignore
    assert machine.models[0] is machine
    assert len(machine.models) == 1


def test_restore_from_markup() -> None:
    machine = HierarchicalMarkupMachine(model=None, **CONFIG)  # type: ignore
    assert isinstance(machine.markup, dict)
    markup = deepcopy(machine.markup)
    del markup['models']  # type: ignore
    rebuild = HierarchicalMarkupMachine(model=None, **markup)  # type: ignore
    assert rebuild.markup == machine.markup


@pytest.mark.skip
def test_graph(tmp_path: Path) -> None:  # pragma: no cover
    FILE_NAME = 'states.png'
    path = tmp_path / FILE_NAME
    # print(f'Saving the state diagram to {path}...')
    machine = HierarchicalAsyncGraphMachine(model=SELF_LITERAL, **CONFIG)  # type: ignore
    machine.get_graph().draw(path, prog='dot')


STATE_MAP = {
    'created': {
        'start': {'dest': 'off'},
    },
    'off': {
        'turn_on': {'dest': 'auto_waiting'},
    },
    'auto_waiting': {
        'turn_off': {'dest': 'off', 'before': 'cancel_task'},
        'on_initialized': {'dest': 'auto_pulling'},
        'on_finished': {'dest': 'auto_pulling'},
        'on_raised': {'dest': 'off'},
    },
    'auto_pulling': {
        'run': {'dest': 'auto_running'},
        'turn_off': {'dest': 'off', 'before': 'cancel_task'},
        'on_raised': {'dest': 'off'},
    },
    'auto_running': {
        'on_finished': {'dest': 'auto_pulling'},
        'turn_off': {'dest': 'off', 'before': 'cancel_task'},
        'on_raised': {'dest': 'off'},
    },
}

TRIGGERS = list({trigger for v in STATE_MAP.values() for trigger in v.keys()})


@settings(max_examples=200)
@given(triggers=st.lists(st.sampled_from(TRIGGERS)))
async def test_transitions(triggers: list[str]) -> None:
    machine = HierarchicalAsyncMachine(model=SELF_LITERAL, **CONFIG)  # type: ignore
    assert machine.is_created()

    for trigger in triggers:
        prev = machine.state
        if (map_ := STATE_MAP[prev].get(trigger)) is None:
            await getattr(machine, trigger)()
            assert machine.state == prev
            continue

        if before := map_.get('before'):
            setattr(machine, before, AsyncMock())

        assert await getattr(machine, trigger)() is True
        dest = map_['dest']
        assert getattr(machine, f'is_{dest}')()

        if before:
            assert getattr(machine, before).call_count == 1
            assert getattr(machine, before).await_count == 1



import asyncio
import time
from collections.abc import Callable

from nextline import Nextline
from nextline_schedule.auto import AutoMode

from .funcs import pull_func_factory


def f() -> None:  # pragma: no cover
    time.sleep(0.001)


async def pull() -> Callable[[], None]:  # pragma: no cover
    return f


async def test_turn_off_while_waiting() -> None:
    nextline = Nextline(
        statement=f,
        trace_threads=True,
        trace_modules=True,
        timeout_on_exit=60,
    )
    queue = pull_func_factory('queue')
    auto_mode = AutoMode(nextline=nextline, scheduler=pull, queue=queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await nextline.run()
            async for prompt in nextline.prompts():  # pragma: no branch
                break
            await auto_mode.turn_on()
            await auto_mode.turn_off()
            await nextline.send_pdb_command(
                command='continue',
                prompt_no=prompt.prompt_no,
                trace_no=prompt.trace_no,
            )

    expected = ['off', 'auto_waiting', 'off']
    assert expected == await states


async def subscribe_state(auto_mode: AutoMode) -> list[str]:
    return [state async for state in auto_mode.subscribe_state()]



import asyncio

from hypothesis import given, settings
from hypothesis import strategies as st

from nextline import Nextline
from nextline_schedule.auto import AutoMode

from .funcs import STATEMENT_TEMPLATE, pull_func_factory, until_state


async def test_pull_func_factory() -> None:
    name = 'script'
    pull = pull_func_factory(name)

    statement = await pull()
    assert isinstance(statement, str)
    assert name in statement
    assert '1' in statement


@settings(deadline=None)
@given(data=st.data())
async def test_until_state(data: st.DataObject) -> None:
    state = data.draw(st.sampled_from(['finished', 'running']))
    count = data.draw(st.integers(min_value=1, max_value=3))

    statement = STATEMENT_TEMPLATE.format(name='init', count=1)
    nextline = Nextline(statement=statement)
    scheduler = pull_func_factory('schedule')
    queue = pull_func_factory('queue')
    auto_mode = AutoMode(nextline=nextline, scheduler=scheduler, queue=queue)

    async with auto_mode:
        async with nextline:
            await auto_mode.turn_on()
            await until_state(nextline, state=state, count=count)

            # TODO: Without the sleep, the exception
            # `asyncio.exceptions.CancelledError` is caught at
            # https://github.com/simonsobs/nextline/blob/v0.7.18/nextline/fsm/callback.py#L65
            # await asyncio.sleep(0.0001)

            await auto_mode.turn_off()
            await until_state(nextline)



import pytest
from strawberry import Schema

from nextline_schedule.schema import Mutation, Query, Subscription


@pytest.fixture(scope='session')
def schema() -> Schema:
    '''GraphQL schema

    The scope is `session` because Hypothesis doesn't allow function-scoped
    fixtures. This is fine as the schema is stateless.
    '''
    return Schema(query=Query, mutation=Mutation, subscription=Subscription)



import asyncio

from nextline import Nextline
from nextline_schedule.auto import AutoMode

from .funcs import STATEMENT_TEMPLATE, pull_func_factory, until_state


async def test_one() -> None:
    statement = STATEMENT_TEMPLATE.format(name='init', count=1)
    nextline = Nextline(statement=statement)
    scheduler = pull_func_factory('schedule')
    queue = pull_func_factory('queue')
    auto_mode = AutoMode(nextline=nextline, scheduler=scheduler, queue=queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            assert auto_mode.state == 'off'
            await auto_mode.turn_on()
            await asyncio.sleep(0.0001)
            await until_state(nextline, state='running', count=2)
            await asyncio.sleep(0.0001)
            await auto_mode.turn_off()

    expected = [
        'off',
        'auto_waiting',
        'auto_pulling',
        'auto_running',
        'auto_pulling',
        'auto_running',
        'off',
    ]
    assert expected == await states


async def subscribe_state(auto_mode: AutoMode) -> list[str]:
    return [state async for state in auto_mode.subscribe_state()]



'''The configuration of the finite state machine of the auto mode states.

The package "transitions" is used: https://github.com/pytransitions/transitions

State Diagram:

                     .-------------.
                     |   Created   |
                     '-------------'
                            | start()
                            |
                            v
                     .-------------.
     .-------------->|     Off     |<--------------.
     |               '-------------'               |
     |                      | turn_on()            |
   turn_off()               |                 on_raised()
     |                      |                      |
     |                      |                      |
     |   .------------------+------------------.   |
     |   |   Auto           |                  |   |
     |   |                  v                  |   |
     |   |           .-------------.           |   |
     |   |           |   Waiting   |           |   |
     |   |           '-------------'           |   |
     |   |                  | on_initialized() |   |
     |   |                  | on_finished()    |   |
     |   |                  v                  |   |
     |   |           .-------------.           |   |
     |   |           |   Pulling   |           |   |
     '---|           '-------------'           |---'
         |         run() |     ^               |
         |               |     |               |
         |               v     | on_finished() |
         |           .-------------.           |
         |           |   Running   |           |
         |           '-------------'           |
         |                                     |
         '-------------------------------------'

>>> class Model:
...     def on_enter_auto_waiting(self):
...         print('enter the waiting state')
...         self.on_finished()
...
...     def on_exit_auto_waiting(self):
...        print('exit the waiting state')
...
...     def on_enter_auto_pulling(self):
...        print('enter the pulling state')

>>> from transitions.extensions import HierarchicalMachine

>>> model = Model()
>>> machine = HierarchicalMachine(model=model, **CONFIG)
>>> model.state
'created'

>>> _ = model.start()
>>> model.state
'off'

>>> _ = model.turn_on()
enter the waiting state
exit the waiting state
enter the pulling state

>>> model.state
'auto_pulling'
'''


_AUTO_SUB_STATE_CONFIG = {
    'name': 'auto',
    'children': ['waiting', 'pulling', 'running'],
    'initial': 'waiting',
    'transitions': [
        ['on_initialized', 'waiting', 'pulling'],
        ['on_finished', 'waiting', 'pulling'],
        ['run', 'pulling', 'running'],
        ['on_finished', 'running', 'pulling'],
    ],
}

CONFIG = {
    'name': 'global',
    'states': ['created', 'off', _AUTO_SUB_STATE_CONFIG],
    'transitions': [
        ['start', 'created', 'off'],
        ['turn_on', 'off', 'auto'],
        ['on_raised', 'auto', 'off'],
        {
            'trigger': 'turn_off',
            'source': 'auto',
            'dest': 'off',
            'before': 'cancel_task',
        },
    ],
    'initial': 'created',
    'queued': True,
    'ignore_invalid_triggers': True,
}



from string import ascii_letters
from unittest.mock import Mock

from hypothesis import given
from hypothesis import strategies as st

from nextline.utils import to_aiter
from nextline_schedule.auto import AutoMode
from nextline_schedule.graphql import SUBSCRIBE_AUTO_MODE_STATE
from tests.schema.conftest import Schema


@given(states=st.lists(st.text(ascii_letters, min_size=1)))
async def test_schema(states: list[str], schema: Schema) -> None:
    if not states:
        # TODO: Test fails for empty list
        return

    auto_mode = Mock(spec=AutoMode)
    auto_mode.subscribe_state = Mock(return_value=to_aiter(states))

    context_auto_mode = {'auto_mode': auto_mode}
    context = {'schedule': context_auto_mode}

    sub = await schema.subscribe(SUBSCRIBE_AUTO_MODE_STATE, context_value=context)

    assert hasattr(sub, '__aiter__')
    idx = 0
    async for resp in sub:
        assert not resp.errors
        assert (data := resp.data)
        assert data['scheduleAutoModeState'] == states[idx]
        idx += 1

    assert idx == len(states)



import strawberry

from .auto import MutationScheduleAutoMode
from .queue import MutationScheduleQueue
from .scheduler import MutationScheduleScheduler


@strawberry.type
class MutationSchedule:
    @strawberry.mutation
    def auto_mode(self) -> MutationScheduleAutoMode:
        return MutationScheduleAutoMode()

    @strawberry.mutation
    def scheduler(self) -> MutationScheduleScheduler:
        return MutationScheduleScheduler()

    @strawberry.mutation
    def queue(self) -> MutationScheduleQueue:
        return MutationScheduleQueue()


@strawberry.type
class Mutation:
    @strawberry.mutation
    def schedule(self) -> MutationSchedule:
        return MutationSchedule()



import datetime
import json
import unittest.mock as mock

import httpx
import pytest
from pytest_httpx import HTTPXMock

from nextline_schedule.scheduler import Scheduler
from nextline_schedule.utils.utc import is_timezone_aware

MOCK_URL = 'https://scheduler-uobd.onrender.com/api/v1/schedule/'

MOCK_UTCNOW = datetime.datetime(2023, 3, 31, 16, 12, 59)
assert not is_timezone_aware(MOCK_UTCNOW)

COMMANDS = '''
import time
# 2023-03-31 16:12:59
time.sleep(7.61)
time.sleep(5.45)
time.sleep(11.82)
time.sleep(35.12)
'''.lstrip()


async def test_success(httpx_mock: HTTPXMock) -> None:
    data = {
        'commands': COMMANDS,
        # 'message': 'Success',
        # 'status': 'ok',
    }
    httpx_mock.add_response(json=data)
    func = Scheduler(api_url=MOCK_URL, length_minutes=1, policy='dummy')

    statement = await func()
    assert statement == COMMANDS

    assert (request := httpx_mock.get_request())

    assert request.method == 'POST'
    assert request.url == MOCK_URL

    expected_content = {
        't0': '2023-03-31 16:12',
        't1': '2023-03-31 16:13',
        'policy': 'dummy',
    }
    assert json.loads(request.content) == expected_content


async def test_400(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(status_code=400)
    func = Scheduler(api_url=MOCK_URL, length_minutes=1, policy='dummy')

    with pytest.raises(RuntimeError):
        await func()


async def test_timeout(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_exception(httpx.ReadTimeout('timeout'))
    func = Scheduler(api_url=MOCK_URL, length_minutes=1, policy='dummy')

    with pytest.raises(httpx.TimeoutException):
        await func()


@pytest.mark.skip
async def test_with_dummy_server() -> None:  # pragma: no cover
    # https://simonsobs.slack.com/archives/D03TBRKRRQ9/p1689280148455099
    test_api_url = 'https://scheduler-uobd.onrender.com/api/v1/schedule/'
    func = Scheduler(api_url=test_api_url, length_minutes=1, policy='dummy')

    statement = await func()
    print(statement)


@pytest.fixture(autouse=True)
def mock_datetime_utcnow(monkeypatch: pytest.MonkeyPatch) -> mock.Mock:
    from nextline_schedule import scheduler

    m = mock.Mock(wraps=datetime)
    m.datetime.utcnow.return_value = MOCK_UTCNOW
    monkeypatch.setattr(scheduler, 'datetime', m)

    return m



from typing import TypedDict
from unittest.mock import AsyncMock

from deepdiff import DeepDiff
from hypothesis import given
from hypothesis import strategies as st

from nextline import Nextline
from nextline_schedule.graphql import MUTATE_SCHEDULER_LOAD_SCRIPT
from nextline_schedule.scheduler import Scheduler
from nextline_test_utils.strategies import st_python_scripts
from tests.schema.conftest import Schema


class Input(TypedDict):
    apiUrl: str | None
    lengthMinutes: int | None
    policy: str | None


@given(d=st.data())
async def test_schema(d: st.DataObject, schema: Schema) -> None:
    init_statement = d.draw(st_python_scripts())
    statement = d.draw(st_python_scripts())

    nextline = Nextline(statement=init_statement)

    scheduler = AsyncMock(spec=Scheduler)
    scheduler.return_value = statement

    async with nextline:
        context_values = {'nextline': nextline, 'schedule': {'scheduler': scheduler}}
        variable_values = {'input': input}

        resp = await schema.execute(
            MUTATE_SCHEDULER_LOAD_SCRIPT,
            context_value=context_values,
            variable_values=variable_values,
        )

        assert (data := resp.data)
        expected_data = {'schedule': {'scheduler': {'loadScript': True}}}
        assert DeepDiff(data, expected_data) == {}

        scheduler.assert_awaited_once()
        assert statement == nextline.statement



__all__ = [
    'PushArg',
    'QueueItem',
    'Queue',
    'QueueEmpty',
    'QueueImp',
]

from .item import PushArg, QueueItem
from .queue import Queue, QueueEmpty
from .queue_imp import QueueImp



import asyncio
from itertools import count

from nextline import Nextline
from nextline_schedule.auto.types import PullFunc

STATEMENT_TEMPLATE = '''
"""{name}: {count}"""
import time
time.sleep(0.01)
'''


def pull_func_factory(name: str) -> PullFunc:
    counter = count(1).__next__

    async def pull_func() -> str:
        await asyncio.sleep(0)
        return STATEMENT_TEMPLATE.format(name=name, count=counter())

    return pull_func


async def until_state(
    nextline: Nextline, state: str = 'finished', count: int = 1
) -> None:
    async for s in nextline.subscribe_state():  # pragma: no branch
        if s != state:
            continue
        count -= 1
        if count == 0:
            break



from unittest.mock import Mock

from nextline_schedule.auto import AutoMode
from nextline_schedule.graphql import QUERY_AUTO_MODE_MODE
from tests.schema.conftest import Schema


async def test_schema(schema: Schema) -> None:
    auto_mode = Mock(spec=AutoMode)
    mode = 'off'
    auto_mode.mode = mode
    context_auto_mode = {'auto_mode': auto_mode}
    context = {'schedule': context_auto_mode}
    resp = await schema.execute(QUERY_AUTO_MODE_MODE, context_value=context)
    assert (data := resp.data)
    expected = {'schedule': {'autoMode': {'mode': mode}}}
    assert data == expected



import strawberry
from strawberry.types import Info

from nextline_schedule.scheduler import Scheduler


def query_scheduler_api_url(info: Info) -> str:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, Scheduler)
    return scheduler._api_url


def query_scheduler_length_minutes(info: Info) -> int:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, Scheduler)
    return scheduler._length_minutes


def query_scheduler_policy(info: Info) -> str:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, Scheduler)
    return scheduler._policy


@strawberry.type
class ScheduleSchedulerPreviewItem:
    script: str
    # TODO: Add metadata, e.g., timestamp, etc.


async def query_scheduler_preview(info: Info) -> ScheduleSchedulerPreviewItem:
    scheduler = info.context['schedule']['scheduler']
    assert callable(scheduler)
    statement = await scheduler()
    assert isinstance(statement, str)
    return ScheduleSchedulerPreviewItem(script=statement)


@strawberry.type
class QueryScheduleScheduler:
    api_url: str = strawberry.field(resolver=query_scheduler_api_url)
    length_minutes: int = strawberry.field(resolver=query_scheduler_length_minutes)
    policy: str = strawberry.field(resolver=query_scheduler_policy)
    preview: ScheduleSchedulerPreviewItem = strawberry.field(
        resolver=query_scheduler_preview
    )



from unittest.mock import Mock

from deepdiff import DeepDiff
from hypothesis import given, settings
from hypothesis import strategies as st

from nextline.utils import to_aiter
from nextline_schedule.graphql import SUBSCRIBE_QUEUE_ITEMS
from nextline_schedule.queue.strategies import Queue, QueueItem, st_queue_item_list
from tests.schema.conftest import Schema


@settings(deadline=None)
@given(item_lists=st.lists(st_queue_item_list(max_size=5), max_size=10))
async def test_schema(item_lists: list[list[QueueItem]], schema: Schema) -> None:
    if not item_lists:
        # TODO: Test fails for empty list
        return

    queue = Mock(spec=Queue)
    queue.subscribe = Mock(return_value=to_aiter(item_lists))

    context_schedule = {'queue': queue}
    context = {'schedule': context_schedule}

    sub = await schema.subscribe(SUBSCRIBE_QUEUE_ITEMS, context_value=context)

    assert hasattr(sub, '__aiter__')
    idx = 0
    async for resp in sub:
        assert not resp.errors
        assert (data := resp.data)
        expected = [
            {
                'name': item.name,
                'script': item.script,
                'id': item.id,
                'createdAt': item.created_at.isoformat(),
            }
            for item in item_lists[idx]
        ]
        actual = data['scheduleQueueItems']
        assert DeepDiff(actual, expected) == {}

        idx += 1

    assert idx == len(item_lists)



from typing import TypedDict
from unittest.mock import Mock

from deepdiff import DeepDiff
from hypothesis import Phase, given, settings
from hypothesis import strategies as st

from nextline_schedule.graphql import MUTATE_QUEUE_PUSH
from nextline_schedule.queue import PushArg, Queue, QueueItem
from nextline_schedule.queue.strategies import st_push_arg
from nextline_test_utils.strategies import st_datetimes, st_graphql_ints
from tests.schema.conftest import Schema


class Input(TypedDict):
    name: str
    script: str


@st.composite
def st_input(draw: st.DrawFn) -> Input:
    push_arg = draw(st_push_arg())
    return {'name': push_arg.name, 'script': push_arg.script}


@settings(phases=(Phase.generate,))  # To avoid shrinking
@given(d=st.data())
async def test_schema(d: st.DataObject, schema: Schema) -> None:
    input = d.draw(st_input())
    item = QueueItem(
        **input,
        id=d.draw(st_graphql_ints(min_value=1)),
        created_at=d.draw(st_datetimes()),
    )

    queue = Mock(spec=Queue)
    queue.push.return_value = item

    context_values = {'schedule': {'queue': queue}}
    variable_values = {'input': input}

    resp = await schema.execute(
        MUTATE_QUEUE_PUSH,
        context_value=context_values,
        variable_values=variable_values,
    )

    assert (data := resp.data)
    expected_data = {
        'schedule': {
            'queue': {
                'push': {
                    'name': item.name,
                    'script': item.script,
                    'id': item.id,
                    'createdAt': item.created_at.isoformat(),
                }
            }
        }
    }
    assert DeepDiff(data, expected_data) == {}

    queue.push.assert_awaited_once_with(PushArg(**input))



from hypothesis import given
from strawberry.types import ExecutionResult

from nextline_schedule.graphql import QUERY_QUEUE_ITEMS
from nextline_schedule.queue.strategies import Queue, st_queue
from tests.schema.conftest import Schema


@given(queue=st_queue())
async def test_schema(queue: Queue, schema: Schema) -> None:
    async with queue:
        context_schedule = {'queue': queue}
        context = {'schedule': context_schedule}
        resp = await schema.execute(QUERY_QUEUE_ITEMS, context_value=context)
        assert isinstance(resp, ExecutionResult)
        assert resp.data
        assert len(resp.data['schedule']['queue']['items']) == len(queue.items)



from string import ascii_lowercase
from unittest.mock import Mock

from deepdiff import DeepDiff
from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.auto import AutoMode
from nextline_schedule.graphql import MUTATE_AUTO_MODE_CHANGE_MODE
from tests.schema.conftest import Schema

MODES = ['off', 'scheduler', 'queue']


@given(mode=st.one_of(st.sampled_from(MODES), st.text(ascii_lowercase, min_size=1)))
async def test_schema(mode: str, schema: Schema) -> None:
    valid = mode in MODES
    auto_mode = Mock(spec=AutoMode)
    context_auto_mode = {'auto_mode': auto_mode}
    context_value = {'schedule': context_auto_mode}
    variable_values = {'mode': mode}
    resp = await schema.execute(
        MUTATE_AUTO_MODE_CHANGE_MODE,
        context_value=context_value,
        variable_values=variable_values,
    )
    assert (data := resp.data)
    expected = {'schedule': {'autoMode': {'changeMode': valid}}}
    assert DeepDiff(data, expected) == {}
    if valid:
        auto_mode.change_mode.assert_awaited_once_with(mode)



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

import pytest
from hypothesis import Phase, given, settings
from hypothesis import strategies as st

from nextline_schedule.queue import QueueEmpty, QueueItem
from nextline_schedule.queue.strategies import st_push_arg, st_queue


class StatefulTest:
    def __init__(self, data: st.DataObject) -> None:
        self._draw = data.draw
        self._queue = data.draw(st_queue())
        self._expected = list[list[QueueItem]]()

    @asynccontextmanager
    async def context(self) -> AsyncIterator[None]:
        await asyncio.sleep(0)
        self._ids = [i.id for i in self._queue.items]
        assert len(self._ids) == len(self._queue)
        yield

    async def _subscribe(self) -> list[list[QueueItem]]:
        return [i async for i in self._queue.subscribe()]

    def _on_expected_publish(self) -> None:
        self._expected.append(list(self._queue.items))

    async def call(self) -> None:
        if not self._ids:
            with pytest.raises(QueueEmpty):
                await self._queue()
            ids = [i.id for i in self._queue.items]
            assert ids == self._ids
        else:
            expected = self._queue.items[0].script
            script = await self._queue()
            assert script == expected
            ids = [i.id for i in self._queue.items]
            assert ids == self._ids[1:]
            self._on_expected_publish()

    async def pop(self) -> None:
        if not self._ids:
            assert await self._queue.pop() is None
            ids = [i.id for i in self._queue.items]
            assert ids == self._ids
        else:
            item = await self._queue.pop()
            assert item is not None
            assert item.id == self._ids[0]
            ids = [i.id for i in self._queue.items]
            assert ids == self._ids[1:]
            self._on_expected_publish()

    async def push(self) -> None:
        arg = self._draw(st_push_arg())
        item = await self._queue.push(arg)
        assert item.name == arg.name
        id = item.id
        ids = [i.id for i in self._queue.items]
        assert ids[-1] == id
        assert ids[:-1] == self._ids
        self._on_expected_publish()

    async def remove(self) -> None:
        fake_id = max(self._ids, default=0) + 1
        id = self._draw(st.sampled_from([*self._ids, fake_id]))
        success = await self._queue.remove(id)
        ids = [i.id for i in self._queue.items]
        if id in self._ids:
            assert success
            idx = self._ids.index(id)
            assert [*ids[:idx], id, *ids[idx:]] == self._ids
            self._on_expected_publish()
        else:
            assert not success
            assert ids == self._ids

    async def move_to_first(self) -> None:
        fake_id = max(self._ids, default=0) + 1
        id = self._draw(st.sampled_from([*self._ids, fake_id]))
        success = await self._queue.move_to_first(id)
        ids = [i.id for i in self._queue.items]
        if id in self._ids:
            assert success
            idx = self._ids.index(id)
            assert ids[0] == id
            assert [*ids[1 : idx + 1], ids[0], *ids[idx + 1 :]] == self._ids
            self._on_expected_publish()
        else:
            assert not success
            assert ids == self._ids

    async def move_to_last(self) -> None:
        fake_id = max(self._ids, default=0) + 1
        id = self._draw(st.sampled_from([*self._ids, fake_id]))
        success = await self._queue.move_to_last(id)
        ids = [i.id for i in self._queue.items]
        if id in self._ids:
            assert success
            idx = self._ids.index(id)
            assert ids[-1] == id
            assert [*ids[:idx], ids[-1], *ids[idx:-1]] == self._ids
            self._on_expected_publish()
        else:
            assert not success
            assert ids == self._ids

    async def move_one_forward(self) -> None:
        fake_id = max(self._ids, default=0) + 1
        id = self._draw(st.sampled_from([*self._ids, fake_id]))
        success = await self._queue.move_one_forward(id)
        ids = [i.id for i in self._queue.items]
        idx = None if id not in self._ids else self._ids.index(id)
        if idx:  # >= 1
            assert success
            assert ids[idx - 1] == id
            assert [
                *ids[: idx - 1],
                ids[idx],
                ids[idx - 1],
                *ids[idx + 1 :],
            ] == self._ids
            self._on_expected_publish()
        else:
            assert not success
            assert ids == self._ids

    async def move_one_backward(self) -> None:
        fake_id = max(self._ids, default=0) + 1
        id = self._draw(st.sampled_from([*self._ids, fake_id]))
        success = await self._queue.move_one_backward(id)
        ids = [i.id for i in self._queue.items]
        idx = None if id not in self._ids else self._ids.index(id)
        if idx is not None and idx != len(self._ids) - 1:
            assert success
            assert ids[idx + 1] == id
            assert [*ids[:idx], ids[idx + 1], ids[idx], *ids[idx + 2 :]] == self._ids
            self._on_expected_publish()
        else:
            assert not success
            assert ids == self._ids

    async def __aenter__(self) -> 'StatefulTest':
        self._task = asyncio.create_task(self._subscribe())
        await asyncio.sleep(0)
        await self._queue.__aenter__()
        self._on_expected_publish()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self._queue.__aexit__(*args, **kwargs)
        await asyncio.sleep(0)
        actual = await self._task
        assert self._expected == actual


@settings(max_examples=500, phases=(Phase.generate,))
@given(data=st.data())
async def test_property(data: st.DataObject) -> None:
    test = StatefulTest(data)

    METHODS = [
        test.call,
        test.pop,
        test.push,
        test.remove,
        test.move_to_first,
        test.move_to_last,
        test.move_one_forward,
        test.move_one_backward,
    ]

    methods = data.draw(st.lists(st.sampled_from(METHODS)))

    async with test:
        for method in methods:
            async with test.context():
                await method()

    await cancel_extra_tasks()


async def cancel_extra_tasks() -> None:
    # For unknown reasons, the task created by `queue.subscribe()` in the
    # `_subscribe` method is sometimes still running.
    extra_tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    for t in extra_tasks:  # pragma: no cover
        t.cancel()
        try:
            await t
        except asyncio.CancelledError:
            pass



from string import ascii_letters
from unittest.mock import Mock

from hypothesis import given, provisional
from hypothesis import strategies as st

from nextline_schedule.graphql import QUERY_SCHEDULER
from nextline_schedule.scheduler import Scheduler
from nextline_test_utils.strategies import st_graphql_ints
from tests.schema.conftest import Schema


@given(d=st.data())
async def test_schema(d: st.DataObject, schema: Schema) -> None:
    api_url = d.draw(provisional.urls())
    length_minutes = d.draw(st_graphql_ints(min_value=1))
    policy = d.draw(st.text(ascii_letters, min_size=1))

    scheduler = Mock(spec=Scheduler)
    scheduler._api_url = api_url
    scheduler._length_minutes = length_minutes
    scheduler._policy = policy

    context = {'schedule': {'scheduler': scheduler}}
    resp = await schema.execute(QUERY_SCHEDULER, context_value=context)
    assert (data := resp.data)

    expected = {
        'schedule': {
            'scheduler': {
                'apiUrl': api_url,
                'lengthMinutes': length_minutes,
                'policy': policy,
            }
        }
    }
    assert data == expected



from unittest.mock import Mock

from nextline_schedule.auto import AutoMode
from nextline_schedule.graphql import MUTATE_AUTO_MODE_TURN_OFF
from tests.schema.conftest import Schema


async def test_schema(schema: Schema) -> None:
    auto_mode = Mock(spec=AutoMode)
    context_auto_mode = {'auto_mode': auto_mode}
    context_value = {'schedule': context_auto_mode}
    resp = await schema.execute(MUTATE_AUTO_MODE_TURN_OFF, context_value=context_value)
    assert (data := resp.data)
    expected = {'schedule': {'autoMode': {'turnOff': True}}}
    assert data == expected
    auto_mode.turn_off.assert_awaited_once()



import dataclasses
from datetime import datetime


@dataclasses.dataclass
class QueueItem:
    id: int
    name: str
    created_at: datetime
    script: str


@dataclasses.dataclass
class PushArg:
    name: str
    script: str

    def to_queue_item(self, id: int) -> QueueItem:
        return QueueItem(
            id=id,
            name=self.name,
            created_at=datetime.utcnow(),
            script=self.script,
        )
